local charUpdatePattern = string.char(0x85, 0xED, 0x0F, 0x84, 0xFF, 0xFF, 0xFF, 0xFF, 0x8B, 0x45, 0x00,
	0x8B, 0x0D, 0xFF, 0xFF, 0xFF, 0xFF, 0x50, 0xE8)
local charUpdateMask = "xxxx????xxxxx????xx"
local charUpdateOffset = 13

local macroUpdatePattern = string.char(0xFF, 0x15, 0xFF, 0xFF, 0xFF, 0xFF, 0x8B, 0x0D, 0xFF, 0xFF, 0xFF, 0xFF, 0xE8)
local macroUpdateMask = "xx????xx????x"
local macroUpdateOffset = 8

local romMouseRClickFlag = 0x8000
local romMouseLClickFlag = 0x80

function getCharUpdatePattern()
	return charUpdatePattern
end

function getCharUpdateMask()
	return charUpdateMask
end

function getCharUpdateOffset()
	return charUpdateOffset
end

function getMacroUpdatePattern()
	return macroUpdatePattern
end

function getMacroUpdateMask()
	return macroUpdateMask
end

function getMacroUpdateOffset()
	return macroUpdateOffset
end

function checkExecutableCompatible()
	if( findPatternInProcess(getProc(), charUpdatePattern, charUpdateMask,
	addresses.staticpattern_char, 1) == 0 ) then
		return false
	end

	if( findPatternInProcess(getProc(), macroUpdatePattern, macroUpdateMask,
	addresses.staticpattern_macro, 1) == 0 ) then
		return false
	end

	return true
end

-- Ask witch character does the user want to be, from the open windows.
function selectGame(character)
	-- Get list of windows in an array
	local windowList = findWindowList("*", "Radiant Arcana")

	if( #windowList == 0 ) then
		print("You need to run rom first!")
		return 0
	end

	charToUse = {}
	for i = 1, #windowList, 1 do
		local process, playerAddress, nameAddress
	    -- open first window
		process = openProcess(findProcessByWindow(windowList[i]))
		-- read player address
		showWarnings(false)
		if addresses["staticbase_char"] and addresses["charPtr_offset"] and addresses["pawnName_offset"] then
			playerAddress = memoryReadUIntPtr(process, addresses["staticbase_char"], addresses["charPtr_offset"])
			-- read player name
			if( playerAddress ) then
				nameAddress = memoryReadUInt(process, playerAddress + addresses["pawnName_offset"])
			end
		end
		-- store the player name, with window number
		if nameAddress == nil then
		    charToUse[i] = "(RoM window "..i..")"
  		else
			local tmp = memoryReadString(process, nameAddress)
			if tmp == nil then
				charToUse[i] = "(RoM window "..i..")"
			else
				charToUse[i] = tmp
			end
		end
		showWarnings(true)
		closeProcess(process)
	end

	if( character ) then
		for i,v in pairs(charToUse) do
			printf("[DEBUG] char: '%s', win: %s\n", tostring(v), tostring(windowList[i]))
			if( string.lower(v) == string.lower(character) ) then
				return windowList[i]
			end
		end
	end

	windowChoice = 1
	-- wait until enter is released
    while keyPressedLocal(key.VK_RETURN) do
    	yrest(200)
    end

    notShown = true
	if (#windowList > 1) then  -- if theres more than 1 window, ask the player witch character to use
		while not keyPressedLocal(key.VK_RETURN) do
	    	if keyPressedLocal(key.VK_UP) or keyPressedLocal(key.VK_DOWN) or notShown then
	        	notShown = false
	    		if keyPressedLocal(key.VK_DOWN) then
	        		windowChoice = windowChoice + 1
	        		if windowChoice > #charToUse then
	        	    	windowChoice = #charToUse
					end
	    		end
	    		if keyPressedLocal(key.VK_UP) then
	        		windowChoice = windowChoice - 1
	        		if windowChoice < 1 then
	        	    	windowChoice = 1
					end
	    		end

				if keyPressedLocal(key.VK_UP) or keyPressedLocal(key.VK_DOWN) then
					print("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n")
				end

				-- start message
	    		printf("Choose your character that you want to play on:\n")

	    		for i = 1, #charToUse, 1 do
	        		if i == windowChoice then
	    				cprintf(240, "\n"..charToUse[i])
					else
	    				printf("\n"..charToUse[i])
					end
				end
				printf("\n")
				yrest(200)
	    	end
		end
		yrest(200)
	end

	return windowList[windowChoice]
end

function getWin(character)
	if( __WIN == nil ) then
  		__WIN = selectGame(character)
	end

	return __WIN
end

function getProc()
	if( __PROC == nil or not windowValid(__WIN) ) then
		if( __PROC ) then closeProcess(__PROC) end
		__PROC = openProcess( findProcessByWindow(getWin()) )
	end

	return __PROC
end

function keyComboPressed(_key)
	if _key and keyPressedLocal(_key.key) then
		if _key.modifier == nil or keyPressedLocal(_key.modifier) then
			return true
		end
	end
end

function isInGame()
	-- Note: if not in game, addresses.isInGame + 0xBF4 is 1 when at the character selection screen.
	if memoryReadBytePtr(getProc(),addresses.loadingScreenPtr, addresses.loadingScreen_offset) == 0 and
	   memoryReadInt(getProc(), addresses.isInGame) == 1 then
		return true
	else
		return false
	end
end

function isMounted()
	local playerAddress = memoryReadIntPtr(getProc(), addresses.staticbase_char, addresses.charPtr_offset) or 0
	if playerAddress ~= 0 then
		local mountaddress = memoryReadInt(getProc(), playerAddress + addresses.charPtrMounted_offset) or 0
		if mountaddress ~= 0 then
			return true
		end
	end
	return false
end

function distance(x1, z1, y1, x2, z2, y2)
	if type(x1) == "table" and type(z1) == "table" then
        y2 = z1.Y or z1[3]
        z2 = z1.Z or z1[2]
        x2 = z1.X or z1[1]
        y1 = x1.Y or x1[3]
        z1 = x1.Z or x1[2]
        x1 = x1.X or x1[1]
    elseif z2 == nil and y2 == nil then -- assume x1,z1,x2,z2 values (2 dimensional)
		z2 = x2
		x2 = y1
		y1 = nil
	end

	if( x1 == nil or z1 == nil or x2 == nil or z2 == nil ) then
		error("Error: nil value passed to distance()", 2)
	end

	if y1 == nil or y2 == nil then -- 2 dimensional calculation
		return math.sqrt( (z2-z1)*(z2-z1) + (x2-x1)*(x2-x1) )
	else -- 3 dimensional calculation
		return math.sqrt( (z2-z1)*(z2-z1) + (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) )
	end
end

--==  FLY FUNTIONS  ==--

function isFlying()
	local swimoffsets = {addresses.charPtr_offset, addresses.pawnSwim_offset1, addresses.pawnSwim_offset2}
	local swimactive = 4
	if memoryReadUByte(getProc(), addresses.swimAddress) == 0x90 and
		memoryReadIntPtr(getProc(), addresses.staticbase_char, swimoffsets) == swimactive then
		return true
	else
		return false
	end
end

function flyToggle()
	local swimoffsets = {addresses.charPtr_offset, addresses.pawnSwim_offset1, addresses.pawnSwim_offset2}
	local swimactive = 4
	if isFlying() then
		memoryWriteString(getProc(), addresses.swimAddress, string.char(unpack(addresses.swimAddressBytes)))
		return false
	else
		memoryWriteString(getProc(), addresses.swimAddress, string.rep(string.char(0x90),#addresses.swimAddressBytes))
		memoryWriteIntPtr(getProc(), addresses.staticbase_char, swimoffsets, swimactive)
		return true
	end
end

--== WALL FUNCTIONS  ==--

function isWallHacked()
	local var =	 memoryReadFloat(getProc(), addresses.wallAddress)
	var = tonumber(string.sub(var,1,3))
	return var == wallhacked_value
end

function wallToggle()
	if isWallHacked() then
		memoryWriteFloat(getProc(), addresses.wallAddress, wallnormal_value)
		return false
	else
		memoryWriteFloat(getProc(), addresses.wallAddress, wallhacked_value)
		return true
	end
end

--==  SPEED FUNCTIONS  ==--

function isSpeedFast()
	local baseSpeed = memoryReadFloat(getProc(), addresses.baseSpeedAddress + addresses.baseSpeedOffset)
	return getSpeed() > baseSpeed
end

function getSpeed()
	local playerAddress = memoryReadIntPtr(getProc(), addresses.staticbase_char, addresses.charPtr_offset) or 0
	if playerAddress ~= 0 then
		local mountaddress = memoryReadInt(getProc(), playerAddress + addresses.charPtrMounted_offset) or 0
		if mountaddress == 0 then
			return memoryReadFloat(getProc(), playerAddress + addresses.speedOffset)
		else
			return memoryReadFloat(getProc(), mountaddress + addresses.speedOffset)
		end
	end
end

function speed(_speed)
	-- Current base speed, includes buff effects.
	local baseSpeed = memoryReadFloat(getProc(), addresses.baseSpeedAddress + addresses.baseSpeedOffset)

	if _speed == false then
		_speed = baseSpeed
	else
		_speed = baseSpeed * (1+percent_speed_increase/100)
	end

	-- Change the speed.
	local playerAddress = memoryReadIntPtr(getProc(), addresses.staticbase_char, addresses.charPtr_offset) or 0
	if playerAddress ~= 0 then
		local mountaddress = memoryReadInt(getProc(), playerAddress + addresses.charPtrMounted_offset) or 0
		if mountaddress == 0 then
			memoryWriteFloat(getProc(), playerAddress + addresses.speedOffset, _speed)
		else
			memoryWriteFloat(getProc(), mountaddress + addresses.speedOffset, _speed)
		end
	end
end

--==  STEALTH FUNCTION  ==--

function showStealthedMobs()
	local objlist= CObjectList()
	objlist:update()
	for k,obj in pairs(objlist.Objects) do
		if obj.Type == PT_MONSTER and obj.Stealthed then
			obj:setStealth(false)
		end
	end
end

--==  NOFALL FUNCTION   ==--

function isNoFallHacked()
	return memoryReadUByte(getProc(), addresses.noFallAddress) == 0x90
end

function noFallToggle()
	if isNoFallHacked() then
		memoryWriteString(getProc(), addresses.noFallAddress, string.char(unpack(addresses.noFallAddressBytes)))
		return false
	else
		memoryWriteString(getProc(), addresses.noFallAddress, string.rep(string.char(0x90),#addresses.noFallAddressBytes))
		return true
	end
end

-- SUICIDE FUNCTION

function suicide()
	-- Teleport up relative to current position
	teleport("UP",suicide_height)

	-- Wait till death
	local playerAddress = memoryReadIntPtr(getProc(), addresses.staticbase_char, addresses.charPtr_offset) or 0
	if playerAddress ~= 0 then
		local escape = false
		repeat
			yrest(100)
			if keyComboPressed(hotkeys.NOFALL_TOGGLE) then
				escape = true
			elseif escape == true then
				break
			end
		until memoryReadInt(getProc(), playerAddress + addresses.charAlive_offset) > 7
	end
end

-- TELEPORT FUNCTION

function teleport(dir, stepSize)
	local proc = getProc()

	if stepSize == nil then
		stepSize = teleport_step_size
	end

	-- Get player coords
	local playerAddress = memoryReadIntPtr(proc, addresses.staticbase_char, addresses.charPtr_offset) or 0
	if playerAddress == 0 then
		return
	else
		local mountaddress = memoryReadInt(getProc(), playerAddress + addresses.charPtrMounted_offset) or 0
		if mountaddress == 0 then
			playerAddress = playerAddress + 4
		else
			playerAddress = mountaddress + 4
		end
	end

	local playerX = memoryReadFloatPtr(proc, playerAddress , addresses.playerX_offset)
	local playerY = memoryReadFloatPtr(proc, playerAddress , addresses.playerY_offset)
	local playerZ = memoryReadFloatPtr(proc, playerAddress , addresses.playerZ_offset)

	if dir == "UP" or dir == "DOWN" then
		-- Calculate new Y coordinate
		local newY
		if dir == "UP" then
			newY = playerY + stepSize
		elseif dir == "DOWN" then
			newY = playerY - stepSize
		end
		-- Write new Y coordinate
		memoryWriteFloatPtr(proc, playerAddress , addresses.playerY_offset, newY) -- y value
	elseif dir == "LEFT" or dir == "RIGHT" or dir =="BACK" or dir == "FORWARD" then
		-- Get camera coords
		local cameraAddress = memoryReadUIntPtr(proc, addresses.staticbase_char, addresses.camPtr_offset) or 0
		local cameraX = memoryReadFloat(proc, cameraAddress + addresses.camX_offset)
		local cameraZ = memoryReadFloat(proc, cameraAddress + addresses.camZ_offset)
		local ratio = math.abs(stepSize/distance(playerX, playerZ, cameraX, cameraZ))
		-- Calculate new X and Z coordinates
		local newX, newZ
		if dir == "LEFT" then
			newX = playerX-(playerZ-cameraZ)*ratio
			newZ = playerZ+(playerX-cameraX)*ratio
		elseif dir == "RIGHT" then
			newX = playerX+(playerZ-cameraZ)*ratio
			newZ = playerZ-(playerX-cameraX)*ratio
		elseif dir == "BACK" then
			newX = playerX-(playerX-cameraX)*ratio
			newZ = playerZ-(playerZ-cameraZ)*ratio
		elseif dir == "FORWARD" then
			newX = playerX+(playerX-cameraX)*ratio
			newZ = playerZ+(playerZ-cameraZ)*ratio
		end
		-- Write new X and Z coordinates
		memoryWriteFloatPtr(proc, playerAddress , addresses.playerX_offset, newX) -- x value
		memoryWriteFloatPtr(proc, playerAddress , addresses.playerZ_offset, newZ) -- z value
	end
end

-- Size function

function isMountSizeHacked()
	local playerAddress = memoryReadIntPtr(getProc(), addresses.staticbase_char, addresses.charPtr_offset) or 0
	if playerAddress ~= 0 then
		local mountaddress = memoryReadInt(getProc(), playerAddress + addresses.charPtrMounted_offset) or 0
		if mountaddress ~= 0 then
			local size = memoryReadFloat(getProc(), mountaddress + addresses.sizeOffset)
			return size ~= 1
		end
	end
	return false
end

function setMountSize(size)
	local playerAddress = memoryReadIntPtr(getProc(), addresses.staticbase_char, addresses.charPtr_offset) or 0
	if playerAddress ~= 0 then
		local mountaddress = memoryReadInt(getProc(), playerAddress + addresses.charPtrMounted_offset) or 0
		if mountaddress ~= 0 then
			memoryWriteFloat(getProc(), mountaddress + addresses.sizeOffset, size)
		end
	end
end

function mountSizeToggle(mountSizeState)
	--if isMountSizeHacked() then -- Already hacked so disable
	--	setMountSize(1)
	--	return false
	--else -- Not hacked
		if mountSizeState == true then
			setMountSize(1)
			return false -- return false but don't set because it's already not hacked
		else -- Hack
			setMountSize(mount_size)
			return true
		end
	--end
end